from typing import Dict, Any, List

from pyhon.parameter.base import HonParameter


def str_to_float(string: str | float) -> float:
    try:
        return int(string)
    except ValueError:
        return float(str(string).replace(",", "."))


class HonParameterRange(HonParameter):
    def __init__(self, key: str, attributes: Dict[str, Any]) -> None:
        super().__init__(key, attributes)
        self._min: float = str_to_float(attributes["minimumValue"])
        self._max: float = str_to_float(attributes["maximumValue"])
        self._step: float = str_to_float(attributes["incrementValue"])
        self._default: float = str_to_float(attributes.get("defaultValue", self._min))
        self._value: float = self._default

    def __repr__(self):
        return f"{self.__class__} (<{self.key}> [{self._min} - {self._max}])"

    @property
    def min(self) -> float:
        return self._min

    @property
    def max(self) -> float:
        return self._max

    @property
    def step(self) -> float:
        return self._step

    @property
    def value(self) -> float:
        return self._value if self._value is not None else self._min

    @value.setter
    def value(self, value: float) -> None:
        value = str_to_float(value)
        if self._min <= value <= self._max and not value % self._step:
            self._value = value
        else:
            raise ValueError(
                f"Allowed: min {self._min} max {self._max} step {self._step}"
            )

    @property
    def values(self) -> List[str]:
        return [str(i) for i in range(int(self.min), int(self.max) + 1, int(self.step))]



from typing import Dict, Any, List

from pyhon.parameter.base import HonParameter


class HonParameterEnum(HonParameter):
    def __init__(self, key: str, attributes: Dict[str, Any]) -> None:
        super().__init__(key, attributes)
        self._default = attributes.get("defaultValue")
        self._value = self._default or "0"
        self._values: List[str] = attributes.get("enumValues", [])

    def __repr__(self) -> str:
        return f"{self.__class__} (<{self.key}> {self.values})"

    @property
    def values(self) -> List[str]:
        return [str(value) for value in self._values]

    @property
    def value(self) -> str | float:
        return self._value if self._value is not None else self.values[0]

    @value.setter
    def value(self, value: str) -> None:
        if value in self.values:
            self._value = value
        else:
            raise ValueError(f"Allowed values {self._value}")



from pyhon.parameter.fixed import HonParameterFixed


class Appliance:
    def data(self, data):
        if data["attributes"]["lastConnEvent"]["category"] == "DISCONNECTED":
            data["attributes"]["parameters"]["machMode"] = "0"
        data["active"] = bool(data.get("attributes", {}).get("activity"))
        data["pause"] = data["attributes"]["parameters"]["machMode"] == "3"
        return data

    def settings(self, settings):
        dry_level = settings["startProgram.dryLevel"]
        if isinstance(dry_level, HonParameterFixed) and dry_level.value == "11":
            settings.pop("startProgram.dryLevel", None)
        return settings



def key_print(data, key="", start=True):
    result = ""
    if isinstance(data, list):
        for i, value in enumerate(data):
            result += key_print(value, key=f"{key}.{i}", start=False)
    elif isinstance(data, dict):
        for k, value in sorted(data.items()):
            result += key_print(value, key=k if start else f"{key}.{k}", start=False)
    else:
        result += f"{key}: {data}\n"
    return result


# yaml.dump() would be done the same, but needs an additional dependency...
def pretty_print(data, key="", intend=0, is_list=False, whitespace="  "):
    result = ""
    if isinstance(data, list):
        if key:
            result += f"{whitespace * intend}{'- ' if is_list else ''}{key}:\n"
            intend += 1
        for i, value in enumerate(data):
            result += pretty_print(
                value, intend=intend, is_list=True, whitespace=whitespace
            )
    elif isinstance(data, dict):
        if key:
            result += f"{whitespace * intend}{'- ' if is_list else ''}{key}:\n"
            intend += 1
        for i, (key, value) in enumerate(sorted(data.items())):
            if is_list and not i:
                result += pretty_print(
                    value, key=key, intend=intend, is_list=True, whitespace=whitespace
                )
            elif is_list:
                result += pretty_print(
                    value, key=key, intend=intend + 1, whitespace=whitespace
                )
            else:
                result += pretty_print(
                    value, key=key, intend=intend, whitespace=whitespace
                )
    else:
        result += f"{whitespace * intend}{'- ' if is_list else ''}{key}{': ' if key else ''}{data}\n"
    return result


def create_command(commands, concat=False):
    result = {}
    for name, command in commands.items():
        if not concat:
            result[name] = {}
        for parameter, data in command.settings.items():
            if data.typology == "enum":
                value = data.values
            elif data.typology == "range":
                value = {"min": data.min, "max": data.max, "step": data.step}
            else:
                continue
            if not concat:
                result[name][parameter] = value
            else:
                result[f"{name}.{parameter}"] = value
    return result



from typing import Optional, Dict, Any, List, TYPE_CHECKING

from pyhon.parameter.base import HonParameter
from pyhon.parameter.enum import HonParameterEnum
from pyhon.parameter.fixed import HonParameterFixed
from pyhon.parameter.program import HonParameterProgram
from pyhon.parameter.range import HonParameterRange

if TYPE_CHECKING:
    from pyhon import HonAPI
    from pyhon.appliance import HonAppliance


class HonCommand:
    def __init__(
        self,
        name: str,
        attributes: Dict[str, Any],
        api: "HonAPI",
        appliance: "HonAppliance",
        programs: Optional[Dict[str, "HonCommand"]] = None,
        program_name: str = "",
    ):
        self._api: HonAPI = api
        self._appliance: "HonAppliance" = appliance
        self._name: str = name
        self._programs: Optional[Dict[str, "HonCommand"]] = programs or {}
        self._program_name: str = program_name
        self._description: str = attributes.get("description", "")
        self._parameters: Dict[str, HonParameter] = self._create_parameters(
            attributes.get("parameters", {})
        )
        self._ancillary_parameters: Dict[str, HonParameter] = self._create_parameters(
            attributes.get("ancillaryParameters", {})
        )

    def __repr__(self) -> str:
        return f"{self._name} command"

    def _create_parameters(self, parameters: Dict) -> Dict[str, HonParameter]:
        result: Dict[str, HonParameter] = {}
        for parameter, attributes in parameters.items():
            if parameter == "zoneMap" and self._appliance.zone:
                attributes["default"] = self._appliance.zone
            match attributes.get("typology"):
                case "range":
                    result[parameter] = HonParameterRange(parameter, attributes)
                case "enum":
                    result[parameter] = HonParameterEnum(parameter, attributes)
                case "fixed":
                    result[parameter] = HonParameterFixed(parameter, attributes)
        if self._programs:
            result["program"] = HonParameterProgram("program", self)
        return result

    @property
    def parameters(self) -> Dict[str, HonParameter]:
        return self._parameters

    @property
    def ancillary_parameters(self) -> Dict[str, HonParameter]:
        return self._ancillary_parameters

    async def send(self) -> bool:
        params = {k: v.value for k, v in self._parameters.items()}
        ancillary_params = {k: v.value for k, v in self._ancillary_parameters.items()}
        return await self._api.send_command(
            self._appliance, self._name, params, ancillary_params
        )

    @property
    def programs(self) -> Dict[str, "HonCommand"]:
        if self._programs is None:
            return {}
        return self._programs

    @property
    def program(self) -> str:
        return self._program_name

    @program.setter
    def program(self, program: str) -> None:
        self._appliance.commands[self._name] = self.programs[program]

    def _get_settings_keys(self, command: Optional["HonCommand"] = None) -> List[str]:
        if command is None:
            command = self
        keys = []
        for key, parameter in (
            command._parameters | command._ancillary_parameters
        ).items():
            if key not in keys:
                keys.append(key)
        return keys

    @property
    def setting_keys(self) -> List[str]:
        if not self._programs:
            return self._get_settings_keys()
        result = [
            key
            for cmd in self._programs.values()
            for key in self._get_settings_keys(cmd)
        ]
        return list(set(result + ["program"]))

    @property
    def settings(self) -> Dict[str, HonParameter]:
        return {
            s: param
            for s in self.setting_keys
            if (param := self._parameters.get(s)) is not None
            or (param := self._ancillary_parameters.get(s)) is not None
        }



import logging
from collections.abc import AsyncIterator
from contextlib import asynccontextmanager
from typing import Optional, Callable, List, Tuple

import aiohttp

from pyhon import const
from pyhon.connection.handler.base import ConnectionHandler

_LOGGER = logging.getLogger(__name__)


class HonAuthConnectionHandler(ConnectionHandler):
    _HEADERS = {"user-agent": const.USER_AGENT}

    def __init__(self, session: Optional[aiohttp.ClientSession] = None) -> None:
        super().__init__(session)
        self._called_urls: List[Tuple[int, str]] = []

    @property
    def called_urls(self) -> List[Tuple[int, str]]:
        return self._called_urls

    @called_urls.setter
    def called_urls(self, called_urls: List[Tuple[int, str]]) -> None:
        self._called_urls = called_urls

    @asynccontextmanager
    async def _intercept(
        self, method: Callable, *args, loop: int = 0, **kwargs
    ) -> AsyncIterator:
        kwargs["headers"] = kwargs.pop("headers", {}) | self._HEADERS
        async with method(*args, **kwargs) as response:
            self._called_urls.append((response.status, response.request_info.url))
            yield response



from typing import Dict, Any, List

from pyhon.parameter.base import HonParameter


class HonParameterFixed(HonParameter):
    def __init__(self, key: str, attributes: Dict[str, Any]) -> None:
        super().__init__(key, attributes)
        self._value = attributes.get("fixedValue", None)

    def __repr__(self) -> str:
        return f"{self.__class__} (<{self.key}> fixed)"

    @property
    def value(self) -> str | float:
        return self._value if self._value is not None else "0"

    @value.setter
    def value(self, value: str | float) -> None:
        # Fixed values seems being not so fixed as thought
        self._value = value

    @property
    def values(self) -> List[str]:
        return list(str(self.value))



class Appliance:
    def data(self, data):
        if data["attributes"]["lastConnEvent"]["category"] == "DISCONNECTED":
            data["attributes"]["parameters"]["temp"] = "0"
            data["attributes"]["parameters"]["onOffStatus"] = "0"
            data["attributes"]["parameters"]["remoteCtrValid"] = "0"
            data["attributes"]["parameters"]["remainingTimeMM"] = "0"

        data["active"] = data["attributes"]["parameters"]["onOffStatus"] == "1"

        return data

    def settings(self, settings):
        return settings



from typing import Dict, Any


class HonParameter:
    def __init__(self, key: str, attributes: Dict[str, Any]) -> None:
        self._key = key
        self._category: str = attributes.get("category", "")
        self._typology: str = attributes.get("typology", "")
        self._mandatory: int = attributes.get("mandatory", 0)
        self._value: str | float = ""

    @property
    def key(self) -> str:
        return self._key

    @property
    def value(self) -> str | float:
        return self._value if self._value is not None else "0"

    @property
    def category(self) -> str:
        return self._category

    @property
    def typology(self) -> str:
        return self._typology

    @property
    def mandatory(self) -> int:
        return self._mandatory



import asyncio
from types import TracebackType
from typing import List, Optional, Dict, Any, Type

from aiohttp import ClientSession
from typing_extensions import Self

from pyhon import HonAPI, exceptions
from pyhon.appliance import HonAppliance


class Hon:
    def __init__(self, email: str, password: str, session: ClientSession | None = None):
        self._email: str = email
        self._password: str = password
        self._session: ClientSession | None = session
        self._appliances: List[HonAppliance] = []
        self._api: Optional[HonAPI] = None

    async def __aenter__(self) -> Self:
        return await self.create()

    async def __aexit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> None:
        await self.close()

    @property
    def api(self) -> HonAPI:
        if self._api is None:
            raise exceptions.NoAuthenticationException
        return self._api

    async def create(self) -> Self:
        self._api = await HonAPI(
            self._email, self._password, session=self._session
        ).create()
        await self.setup()
        return self

    @property
    def appliances(self) -> List[HonAppliance]:
        return self._appliances

    async def _create_appliance(self, appliance_data: Dict[str, Any], zone=0) -> None:
        appliance = HonAppliance(self._api, appliance_data, zone=zone)
        if appliance.mac_address is None:
            return
        await asyncio.gather(
            *[
                appliance.load_attributes(),
                appliance.load_commands(),
                appliance.load_statistics(),
            ]
        )
        self._appliances.append(appliance)

    async def setup(self) -> None:
        appliance: Dict
        for appliance in (await self.api.load_appliances())["payload"]["appliances"]:
            if (zones := int(appliance.get("zone", "0"))) > 1:
                for zone in range(zones):
                    await self._create_appliance(appliance.copy(), zone=zone + 1)
            await self._create_appliance(appliance)

    async def close(self) -> None:
        await self.api.close()



#!/usr/bin/env python
import argparse
import asyncio
import json
import logging
import sys
from getpass import getpass
from pathlib import Path

if __name__ == "__main__":
    sys.path.insert(0, str(Path(__file__).parent.parent))

from pyhon import Hon, HonAPI, helper

_LOGGER = logging.getLogger(__name__)


def get_arguments():
    """Get parsed arguments."""
    parser = argparse.ArgumentParser(description="pyhOn: Command Line Utility")
    parser.add_argument("-u", "--user", help="user for haier hOn account")
    parser.add_argument("-p", "--password", help="password for haier hOn account")
    subparser = parser.add_subparsers(title="commands", metavar="COMMAND")
    keys = subparser.add_parser("keys", help="print as key format")
    keys.add_argument("keys", help="print as key format", action="store_true")
    keys.add_argument("--all", help="print also full keys", action="store_true")
    translate = subparser.add_parser(
        "translate", help="print available translation keys"
    )
    translate.add_argument(
        "translate", help="language (de, en, fr...)", metavar="LANGUAGE"
    )
    translate.add_argument("--json", help="print as json", action="store_true")
    return vars(parser.parse_args())


async def translate(language, json_output=False):
    async with HonAPI(anonymous=True) as hon:
        keys = await hon.translation_keys(language)
    if json_output:
        print(json.dumps(keys, indent=4))
    else:
        clean_keys = (
            json.dumps(keys)
            .replace("\\n", "\\\\n")
            .replace("\\\\r", "")
            .replace("\\r", "")
        )
        keys = json.loads(clean_keys)
        print(helper.pretty_print(keys))


async def main():
    args = get_arguments()
    if language := args.get("translate"):
        await translate(language, json_output=args.get("json"))
        return
    if not (user := args["user"]):
        user = input("User for hOn account: ")
    if not (password := args["password"]):
        password = getpass("Password for hOn account: ")
    async with Hon(user, password) as hon:
        for device in hon.appliances:
            print("=" * 10, device.appliance_type, "-", device.nick_name, "=" * 10)
            if args.get("keys"):
                data = device.data.copy()
                attr = "get" if args.get("all") else "pop"
                print(
                    helper.key_print(
                        data["attributes"].__getattribute__(attr)("parameters")
                    )
                )
                print(helper.key_print(data.__getattribute__(attr)("appliance")))
                print(helper.key_print(data))
                print(
                    helper.pretty_print(
                        helper.create_command(device.commands, concat=True)
                    )
                )
            else:
                print(helper.pretty_print({"data": device.data}))
                print(
                    helper.pretty_print(
                        {"settings": helper.create_command(device.commands)}
                    )
                )


def start():
    try:
        asyncio.run(main())
    except KeyboardInterrupt:
        print("Aborted.")


if __name__ == "__main__":
    start()



import logging
from collections.abc import AsyncIterator
from contextlib import asynccontextmanager
from typing import Callable, Dict

from pyhon import const
from pyhon.connection.handler.base import ConnectionHandler

_LOGGER = logging.getLogger(__name__)


class HonAnonymousConnectionHandler(ConnectionHandler):
    _HEADERS: Dict = ConnectionHandler._HEADERS | {"x-api-key": const.API_KEY}

    @asynccontextmanager
    async def _intercept(self, method: Callable, *args, **kwargs) -> AsyncIterator:
        kwargs["headers"] = kwargs.pop("headers", {}) | self._HEADERS
        async with method(*args, **kwargs) as response:
            if response.status == 403:
                _LOGGER.error("Can't authenticate anymore")
            yield response



import importlib
import logging
from contextlib import suppress
from typing import Optional, Dict, Any
from typing import TYPE_CHECKING

from pyhon import helper
from pyhon.commands import HonCommand
from pyhon.parameter.fixed import HonParameterFixed

if TYPE_CHECKING:
    from pyhon import HonAPI


_LOGGER = logging.getLogger(__name__)


class HonAppliance:
    def __init__(
        self, api: Optional["HonAPI"], info: Dict[str, Any], zone: int = 0
    ) -> None:
        if attributes := info.get("attributes"):
            info["attributes"] = {v["parName"]: v["parValue"] for v in attributes}
        self._info: Dict = info
        self._api: Optional[HonAPI] = api
        self._appliance_model: Dict = {}

        self._commands: Dict = {}
        self._statistics: Dict = {}
        self._attributes: Dict = {}
        self._zone: int = zone

        try:
            self._extra = importlib.import_module(
                f"pyhon.appliances.{self.appliance_type.lower()}"
            ).Appliance()
        except ModuleNotFoundError:
            self._extra = None

    def __getitem__(self, item):
        if self._zone:
            item += f"Z{self._zone}"
        if "." in item:
            result = self.data
            for key in item.split("."):
                if all(k in "0123456789" for k in key) and isinstance(result, list):
                    result = result[int(key)]
                else:
                    result = result[key]
            return result
        if item in self.data:
            return self.data[item]
        if item in self.attributes["parameters"]:
            return self.attributes["parameters"].get(item)
        return self.info[item]

    def get(self, item, default=None):
        try:
            return self[item]
        except (KeyError, IndexError):
            return default

    def _check_name_zone(self, name: str, frontend: bool = True) -> str:
        middle = " Z" if frontend else "_z"
        if (attribute := self._info.get(name, "")) and self._zone:
            return f"{attribute}{middle}{self._zone}"
        return attribute

    @property
    def appliance_model_id(self) -> str:
        return self._info.get("applianceModelId", "")

    @property
    def appliance_type(self) -> str:
        return self._info.get("applianceTypeName", "")

    @property
    def mac_address(self) -> str:
        return self.info.get("macAddress", "")

    @property
    def unique_id(self) -> str:
        return self._check_name_zone("macAddress", frontend=False)

    @property
    def model_name(self) -> str:
        return self._check_name_zone("modelName")

    @property
    def nick_name(self) -> str:
        return self._check_name_zone("nickName")

    @property
    def commands_options(self):
        return self._appliance_model.get("options")

    @property
    def commands(self):
        return self._commands

    @property
    def attributes(self):
        return self._attributes

    @property
    def statistics(self):
        return self._statistics

    @property
    def info(self):
        return self._info

    @property
    def zone(self) -> int:
        return self._zone

    async def _recover_last_command_states(self, commands):
        command_history = await self._api.command_history(self)
        for name, command in commands.items():
            last = next(
                (
                    index
                    for (index, d) in enumerate(command_history)
                    if d.get("command", {}).get("commandName") == name
                ),
                None,
            )
            if last is None:
                continue
            parameters = command_history[last].get("command", {}).get("parameters", {})
            if command.programs and parameters.get("program"):
                command.program = parameters.pop("program").split(".")[-1].lower()
                command = self.commands[name]
            for key, data in command.settings.items():
                if (
                    not isinstance(data, HonParameterFixed)
                    and parameters.get(key) is not None
                ):
                    with suppress(ValueError):
                        data.value = parameters.get(key)

    async def load_commands(self):
        raw = await self._api.load_commands(self)
        self._appliance_model = raw.pop("applianceModel")
        for item in ["settings", "options", "dictionaryId"]:
            raw.pop(item)
        commands = {}
        for command, attr in raw.items():
            if "parameters" in attr:
                commands[command] = HonCommand(command, attr, self._api, self)
            elif "parameters" in attr[list(attr)[0]]:
                multi = {}
                for program, attr2 in attr.items():
                    program = program.split(".")[-1].lower()
                    cmd = HonCommand(
                        command,
                        attr2,
                        self._api,
                        self,
                        programs=multi,
                        program_name=program,
                    )
                    multi[program] = cmd
                    commands[command] = cmd
        self._commands = commands
        await self._recover_last_command_states(commands)

    @property
    def settings(self):
        result = {}
        for name, command in self._commands.items():
            for key, setting in command.settings.items():
                result[f"{name}.{key}"] = setting
        if self._extra:
            return self._extra.settings(result)
        return result

    @property
    def parameters(self):
        result = {}
        for name, command in self._commands.items():
            for key, parameter in (
                command.parameters | command.ancillary_parameters
            ).items():
                result.setdefault(name, {})[key] = parameter.value
        return result

    async def load_attributes(self):
        self._attributes = await self._api.load_attributes(self)
        _LOGGER.warning(self._attributes)
        for name, values in (
            self._attributes.pop("shadow", {}).get("parameters", {}).items()
        ):
            self._attributes.setdefault("parameters", {})[name] = values["parNewVal"]

    async def load_statistics(self):
        self._statistics = await self._api.load_statistics(self)

    async def update(self):
        await self.load_attributes()

    @property
    def data(self):
        result = {
            "attributes": self.attributes,
            "appliance": self.info,
            "statistics": self.statistics,
            **self.parameters,
        }
        if self._extra:
            return self._extra.data(result)
        return result

    @property
    def diagnose(self):
        data = self.data.copy()
        for sensible in ["PK", "SK", "serialNumber", "code", "coords"]:
            data["appliance"].pop(sensible, None)
        result = helper.pretty_print({"data": self.data}, whitespace="\u200B \u200B ")
        result += helper.pretty_print(
            {"commands": helper.create_command(self.commands)},
            whitespace="\u200B \u200B ",
        )
        return result.replace(self.mac_address, "12-34-56-78-90-ab")



class Appliance:
    def data(self, data):
        if data["attributes"]["lastConnEvent"]["category"] == "DISCONNECTED":
            data["attributes"]["parameters"]["machMode"] = "0"
        data["active"] = bool(data.get("attributes", {}).get("activity"))
        data["pause"] = data["attributes"]["parameters"]["machMode"] == "3"
        return data

    def settings(self, settings):
        return settings



from typing import List, TYPE_CHECKING, Dict

from pyhon.parameter.enum import HonParameterEnum

if TYPE_CHECKING:
    from pyhon.commands import HonCommand


class HonParameterProgram(HonParameterEnum):
    _FILTER = ["iot_recipe", "iot_guided"]

    def __init__(self, key: str, command: "HonCommand") -> None:
        super().__init__(key, {})
        self._command = command
        self._value: str = command.program
        self._programs: Dict[str, "HonCommand"] = command.programs
        self._typology: str = "enum"

    @property
    def value(self) -> str | float:
        return self._value

    @value.setter
    def value(self, value: str) -> None:
        if value in self.values:
            self._command.program = value
        else:
            raise ValueError(f"Allowed values {self.values}")

    @property
    def values(self) -> List[str]:
        values = [v for v in self._programs if all(f not in v for f in self._FILTER)]
        return sorted(values)
